# $Id: rcs.pl,v 1.28 2002/03/07 14:34:03 billl Exp $
# ci all files from 2nd arg using the first arg for pre and post comment symbol

package rcs;

# Usage: &nci(STR1,STR2,FILES)
# does check-in for multiple files not previously checked in
# the Id string is flanked by STR1 and STR2
# will create RCS dir if not already there

# new check in
sub nci {
    local($beg,@files) = @_;
    local($flag=1);
    ($beg eq "help") and die "check in new files creating RCS dir and adding '\$Id\$' if not present.
\trcs::nci(comment_string,filenames) eg nci(%%,ctl.tex,text.tex)\n"; 
    if (-f $beg) {unshift(@files,$beg); $flag=0;} # if $beg is a file not a string
    if (! -e "RCS") {mkdir("RCS",0755); print "Creating RCS directory\n";}

    foreach $bb (@files) {
	$mode = 
	open(FL,"<$bb");
	$head = <FL>;
	if ($flag && $head !~ /\$He/ && $head !~ /\$Id/) {
	    print "    ****Adding $beg \$Id\$ header to $bb****\n";
	    open(TP,">$bb.tmp");
	    print TP "$beg\$Id\$\n";
	    close TP;
	    `cat $bb >> $bb.tmp ; mv -f $bb.tmp $bb`;
	}
	print "Checking in $bb\n";
	`ci -u -q -t-"" $bb`;
    }
    chmod 0644,@files;
}

# Usage: &chg(@FILES)
# looks at all the chosen files and checks whether these have been
# changed from the original of that version
# if no files given will look at all of RCS directory
# initial flag 'n' means just give file names
# can't use '-n' as a flag because will be picked up by perl

sub chg {
    ($ARGV[0] eq "help") and die "
    Usage: rcs::chg(FILES)
    looks at all the chosen files and checks whether these have been
    changed from the original of that version
    If no files given will look at all of RCS directory
    initial flag 'n' means just give file names without diffs
    can't use '-n' as a flag because will be picked up by perl.\n\n";
    if ($ARGV[0] eq "n") { $nodif=1; shift; } else { $nodif=0; }
    if ($#_ < 0) { # with no arguments check everything
	opendir(RCS,"RCS");
	@files = grep(/,v$/,readdir(RCS)); # just read ,v files (not . ..)
	foreach (@files) {
	    $_ =~ s/,v//; } # get rid of the ,v part
    } else { @files = @_; }

    foreach $FL (@files) {
	if (-e $FL && -e "RCS/$FL,v") {
	    open FL;
	    $cver = <FL>; chop $cver;
	    close FL;
	    if ($cver =~ /\$Id|\$He/) {
		$cver =~ s/.*,v (1\.[0-9][0-9]*).*/$1/;
	    } else { $cver = ""; } # no header line
	    @diffout = `rcsdiff -w -q -kkvl -r$cver $FL`;
	    while ("@diffout" =~ /\$Id|\$He/) { # eliminate header lines
		shift @diffout; }
	    $diffout = "@diffout";
	    if ($diffout ne "") {
		$diffout =~ s/\n/\n\t/g;
		if ($nodif==0) {
		    print "$FL ".(($cver)?"v$cver":"current version")."\n\t$diffout\n";
		} else { print "$FL "; }
	    }
	}
    }
    print "\n";
}

# Usage: &clean
# gets rid of any file that exists in the rcs directory
# does not check if changes are extant so good to precede with a diffs

sub clean {
    ($ARGV[0] eq "help") and die "rcs::clean removes all of the files that ,v files in RCS dir\n";
    opendir(RCS,"RCS") || die "No RCS directory in $ENV{PWD}.\n";
    @files = grep(/,v$/,readdir(RCS)); # just read ,v files (not . ..)
    foreach (@files) { $_ =~ s/,v//; }
    foreach $FL (@files) {
	print "Removing $FL.\n";
	unlink($FL);
    }
}

# Usage: &head
# prints out the head version of all files
sub head {
    ($ARGV[0] eq "help") and die "rcs::head prints out the top version number for RCS files\n";
    opendir(RCS,"RCS") || die "No RCS directory in $ENV{PWD}.\n";
    if ($#_ < 0) { # with no arguments check everything
	opendir(RCS,"RCS");
	@files = grep(/,v$/,readdir(RCS)); # just read ,v files (not . ..)
	foreach (@files) {
	    $_ =~ s/,v//; } # get rid of the ,v part
    } else { @files = @_; }
    foreach $FL (@files) {
	open (FL,"RCS/$FL"); $_ = <FL>; chop; close FL;
	s/head	[0-9]*\.([0-9][0-9]*)/$1/; chop;
	print "$FL $_ ";
	if (-e $FL) { # there is a copy of this file checked out
	    open FL; $_ = <FL>; chop; close FL;	    
	    if (s/.*,v 1\.([0-9][0-9]*).*/$1/) {
		print "($_)";} else { print "(?)"; } # else can't find version
	}
	print "\n";
    }
}

# Usage: &curr
# prints out the current version of all files

sub curr {
    ($ARGV[0] eq "help") and die "rcs::curr prints out the current version number for all files in cwd\n";
    if ($#_ < 0) {		# with no arguments check everything
	opendir(RCS,"RCS")  || die "No RCS directory in $ENV{PWD}.\n";
	@files = grep(/,v$/,readdir(RCS)); # just read ,v files (not . ..)
	foreach (@files) { $_ =~ s/,v//; } # get rid of the ,v part
    } else { @files = @_; }
    foreach $FL (@files) {
	open(FL,"RCS/$FL,v") || die "Can't open RCS/$FL.\n"; 
	$_ = <FL>; chop; close FL;
	s/head	[0-9]*\.([0-9][0-9]*)/$1/; chop;
	push(@head,$_);
	if (-e $FL) { # there is a copy of this file checked out
	    open FL; $_ = <FL>; chop; close FL;	    
	    if (s/.*,v 1\.([0-9][0-9]*).*/$1/) {
		push(@curl,$_);} else { push(@curl,"?"); }
	} else { push(@curl,"NONE"); }
	printf("%8s\tv%1s\t(%1s)\n",$FL,$curl[$#curl],$head[$#head]);
    }
}

# Usage: &date($date)
# looks at all files in the directory and determines the version
# available on a particular date; then shows changes from then to now
sub date {
    ($ARGV[0] eq "help") and die "
Usage: rcs::date(date), eg 'July 28, 1998'
looks at all files in RCS directory and determines the version
available on a particular date; then shows changes from then to now\n\n";
    $sep = ("*" x 60)."\n";	# a line of *'s
    $sep1 = ("_" x 60)."\n";
    $date = shift @_;
    if ($#_ < 0) {		# with no arguments check everything
	opendir(RCS,"RCS");
	@files = grep(/,v$/,readdir(RCS)); # just read ,v files (not . ..)
	foreach (@files) { $_ =~ s/,v//; } # get rid of the ,v part
    } else { @files = @_; }

    foreach $FL (@files) {
	@text = `rlog -d\"$date\" $FL`;
	open FL;
	$cver = <FL>; chop $cver;
	close FL;
	if ($cver =~ /\$Id|\$He/) {
	    $cver =~ s/.*,v (1\.[0-9][0-9]*).*/$1/;
	} else { $cver = ""; }	# no header line
	if ($text[-4] =~ /revision ([0-9]+\.[0-9]+)/) {
	    $vers{$FL} = "-r$1 -r$cver";
	    print "$FL: currently $cver was $text[-4]$text[-3]$text[-2]$sep1";
	} else {
	    # else was not yet created at that time
	    print "$FL had not yet been created.\n$sep1";
	}
    }
    foreach $FL (sort keys %vers) {
	@diffout = `rcsdiff -w -q -kkvl $vers{$FL} $FL`;
	while ("@diffout" =~ /\$Id|\$He/) { # eliminate header lines
	    shift @diffout; }
	$diffout = "@diffout";
	if ($diffout ne "") {
	    $diffout =~ s/\n/\n\t/g;
	    print "$FL $vers{$FL} ...\n\n\t$diffout\n$sep";
	}
    }
}

# Usage: &dateget "$date" $files (or all files)
# pulls out the files that were being used on the date given
sub dateget {
    ($ARGV[0] eq "help") and die "
Usage: rcs::date(date), eg 'July 28, 1998' or '1998/08/14 19:50' from Id line
looks at all files in RCS directory and checks out the version
available on a particular date.\n\n";
    $date = shift @_;
    if ($#_ < 0) {		# with no arguments check everything
	opendir(RCS,"RCS");
	@files = grep(/,v$/,readdir(RCS)); # just read ,v files (not . ..)
	foreach (@files) { $_ =~ s/,v//; } # get rid of the ,v part
    } else { @files = @_; }

    foreach $FL (@files) {
	@text = `rlog -d\"$date\" $FL`;
	if ($text[-4] =~ /revision ([0-9]+\.[0-9]+)/) {
	    print "$FL: $text[-4]$text[-3]$text[-2]\n";
	    $comm = "co -f -r$1 $FL";
	    system("$comm");
	} else {
	    # else was not yet created at that time
	    print "$FL had not yet been created.\n";
	}
    }
}

sub delta {
    ($ARGV[0] eq "help") and die "delta 'comment' [file1 file2 ...]
  performs rcs lock and ci function on multiple files.\n";
    $| = 1;
    local($flag=1);
    $comm = shift;
    if (! -f $comm and $#_<0) { # with no arguments ci everything
	opendir(RCS,"RCS");
	@files = grep(/,v$/,readdir(RCS)); # just read ,v files (not . ..)
	foreach (@files) { $_ =~ s/,v//; } # get rid of the ,v part
    } else { @files = @_; }
    # if $comm is a file not a comment string:
    if (-f $comm) {unshift(@files,$comm); $comm="CHECK IN @files";} 

    $verexp = ".+\n[0-9]+\.([0-9]+) *locked\n.+";
    $date = `datestring`; chop $date;
    $ind = "";
    -e "RCS" || die "No RCS directory.\n";

    foreach $bb (@files) {
	if (!-f $bb) { print "$bb not present.\n"; next; }
	@stat = stat(_);
	if (!-f "RCS/$bb,v") { print "$bb not under RCS.\n"; next; }
	$ver =    `sh -c 'rcs -l $bb' 2>&1`;
	print "$bb => "; 
	$newver = `sh -c 'ci -u -m"$comm" $bb' 2>&1`;
	if ($newver =~ /unchanged/) {
	    print " NO CHANGE\n";
	} else {
	    $newver =~ s/.+\nnew revision: ([0-9.]+);.+\n.+/$1/;
	    print $newver; chop($newver);
	    $ind .= " $bb ($newver)";
	} 
	chmod $stat[2],$bb;
    }
    if ($ind && (-f "index" || -l "index")) { 
	open(IND,">> index"); select IND; }
    print "$date $ind $comm\n";
}

1;
