;; $Id: hoc.el,v 1.70 2002/03/26 15:21:18 billl Exp $
;;
;; This file is part of GNU Emacs.

;; GNU Emacs is distributed in the hope that it will be useful,
;; but WITHOUT ANY WARRANTY.  No author or distributor
;; accepts responsibility to anyone for the consequences of using it
;; or for whether it serves any particular purpose or works at all,
;; unless he says so in writing.  Refer to the GNU Emacs General Public
;; License for full details.

;; Everyone is granted permission to copy, modify and redistribute
;; GNU Emacs, but only under the conditions described in the
;; GNU Emacs General Public License.   A copy of this license is
;; supposed to have been given to you along with GNU Emacs so you
;; can know your rights and responsibilities.  It should be in a
;; file named COPYING.  Among other things, the copyright notice
;; and this notice must be preserved on all copies.
;; hoc.el was modfied from neuron.el by Gary Holt 16-Mar-1993.
;; 
;; simctrl package by Bill Lytton - 11/1/93
;;

;;  (cond (window-system
;;	   (require 'hilit19)
;;	   (load "hoc_hilit" nil t)))

(defvar hoc-command-hist '("run()") "command history")

(defvar hoc-send-all-sim-buffers nil
  "Send things to all of the simulation buffers instead of just the associated buffer")

(defvar hoc-block-indent 2
  "*Indentation of hoc statements with respect to containing block.")

(defvar hoc-continued-statement-offset 2
  "*Extra indent for lines not starting new statements.")

(defvar hoc-function-start-pattern
  "^[ 	]*\\(proc\\|func\\|iterator\\)[ 	]+\\([a-zA-Z_0-9]+\\)[ 	]*([ 	,a-zA-Z_0-9]*)[ 	]*{"
"A regular expression which matches a hoc procedure or function
definition.")
;; a scratch var for saving the sim-index-buffer when going through
;; rcs-delta-file which calls hoc-mode when reverting file
(defvar hoc-input-chunk-size 256
  "*Send chunks of this size or slightly larger to the process")

(setq hoc-index-buffer nil)
  
;;
;; Keymaps:
;;
(defvar hoc-mode-map nil "Keymap for hoc mode with automatic indentation.")
(defvar hoc-indent-disabled-map nil
  "Keymap for hoc mode with indentation disabled.")

(defun hoc-define-keys (keymap &optional no-indent)
  "Defines standard hoc-mode keys for keymap.  If no-indent is non-nil,
only defines the keys not related to indenting."
  (if no-indent nil
    (define-key keymap "\177" 'backward-delete-char-untabify)
    (define-key keymap "\M-\C-q" 'hoc-indent-exp) ; Indent expression.
    (define-key hoc-mode-map "}"
      '(lambda (arg) (interactive "P")
	 (insert "}")			; Put the brace in.
	 (hoc-indent-line)		; Indent the line properly.
	 (delete-char -1)		; Remove it.
	 (self-insert-command (prefix-numeric-value arg)))) ; Insert it so the
					; matching brace blinks.
  )
  (define-key keymap "\C-x\C-e" 'hoc-send-line) ; Send line to neuron.
  (define-key keymap "\e\t"     'nrn-complete-symbol)
  (define-key keymap [C-return] 'hoc-send-line-and-save) ; Send line to neuron.
  (define-key keymap "\C-xw"    'hoc-send-region); Send region to neuron.
  (define-key keymap "\C-xc"    'hoc-neuron-command); Send command to neuron.
  (define-key keymap "\C-xp"    'hoc-send-proc); Send proc to neuron.
  (define-key keymap "\M-\C-x"  'eval-defun) ; to use lisp in notebook
  (define-key keymap "\C-x\C-n" 'hoc-next-function)
  (define-key keymap "\C-x\C-a" 'hoc-toggle-send-all)
  (define-key keymap "\C-xl"    'sim-load-sim)
  (define-key keymap "\C-x\C-l" 
    '(lambda () "run xopen on current buffer's file" (interactive)
       (sim-open-file (expand-file-name buffer-file-name))))
  (define-key keymap "\C-xf"    'sim-load-file)
  (define-key keymap "\C-xs"    'sim-save-sim)
  (define-key keymap "\C-x "    'sim-rcs-status)
  (define-key keymap "\C-x?"    'sim-template-status)
  (define-key keymap "\C-xz"    'sim-switch-to-buffer)
  (define-key keymap "\C-xd"    'sim-direct-output)
  (define-key keymap "\C-x\C-p" 'hoc-previous-function))
  
(if hoc-indent-disabled-map nil	       
  (setq hoc-indent-disabled-map (make-sparse-keymap))
  (hoc-define-keys hoc-indent-disabled-map t))

(if hoc-mode-map nil
  (setq hoc-mode-map (make-sparse-keymap))
  (hoc-define-keys hoc-mode-map))

(defun hoc-menu-bar (map)
  (define-key map [menu-bar hoc] (cons "Hoc" (make-sparse-keymap "Hoc")))
  (define-key map [menu-bar hoc send-buffer] '("Send Buffer" . hoc-send-buffer))
  (define-key map [menu-bar hoc send-region] '("Send Region" . hoc-send-region))
  (define-key map [menu-bar hoc send-proc] '("Send Proc" . hoc-send-proc))
  (define-key map [menu-bar hoc send-line] '("Send Line" . hoc-send-line))
  (define-key map [menu-bar hoc previous-proc] 
    '("Previous Proc" . hoc-previous-function ))
  (define-key map [menu-bar hoc next-proc] '("Next Proc" . hoc-next-function )))

;;
;; Syntax table:
;; This syntax table was stolen shamelessly from c-mode.el.
;;
(defvar hoc-mode-syntax-table nil
  "Syntax table in use in hoc-mode buffers.")
(if hoc-mode-syntax-table nil
  (setq hoc-mode-syntax-table (make-syntax-table))
  (modify-syntax-entry ?\\ "\\" hoc-mode-syntax-table)
  (modify-syntax-entry ?/ ". 14" hoc-mode-syntax-table)
  (modify-syntax-entry ?* ". 23" hoc-mode-syntax-table)
  (modify-syntax-entry ?+ "." hoc-mode-syntax-table)
  (modify-syntax-entry ?- "." hoc-mode-syntax-table)
  (modify-syntax-entry ?= "." hoc-mode-syntax-table)
  (modify-syntax-entry ?% "." hoc-mode-syntax-table)
  (modify-syntax-entry ?< "." hoc-mode-syntax-table)
  (modify-syntax-entry ?> "." hoc-mode-syntax-table)
  (modify-syntax-entry ?& "." hoc-mode-syntax-table)
  (modify-syntax-entry ?| "." hoc-mode-syntax-table)
  (modify-syntax-entry ?\' "\"" hoc-mode-syntax-table))

;;
;; Turn on hoc mode:
;;
(defun hoc-mode ()
  "Major mode for editing .hoc files for neurons.  

  \\{hoc-mode-map}

Entry into this mode calls the value of hoc-mode-hook with no arguments,
if that value is non-nil.

By default, tab indents the current line according to indentation rules similar
to c-mode.  If you want tab to insert a hardware tab, you can turn off hoc-mode
indentation by calling hoc-indent-toggle.

Variables controlling hoc mode indentation:
 hoc-block-indent
   Additional indentation for each block surrounded by braces.  Only relevant
   if hoc-indent-enable is non-nil.
 hoc-continued-statement-offset
   Additional indentation for a line which is a continuation of the previous
   line.  Only relevant if hoc-indent-enable is non-nil."
  (interactive)
  (let (hfi (sel selective-display))
    (kill-all-local-variables)		; Trash previous major mode.
    (setq major-mode 'hoc-mode)
    (setq mode-name "Hoc")
    (if (setq hfi (hoc-find-index))
	(sim-copy-local-variables sim-localvar-list hfi (current-buffer)))
    (setq sim-redirect-output nil)
    (make-local-variable 'outline-regexp)
    (make-local-variable 'comment-start)
    (make-local-variable 'comment-end)
    (make-local-variable 'comment-start-skip)
    (make-local-variable 'comment-indent-function)
    (make-local-variable 'require-final-newline)
    (make-local-variable 'indent-line-function)
    (make-local-variable 'hoc-indent-enabled)
    (make-local-variable 'rcs-delta-prehook)
    (make-local-variable 'rcs-delta-posthook)
    (make-local-variable 'kill-buffer-hook)
    (setq rcs-delta-prehook '(lambda () (setq rcs-index-buffer sim-index-buffer)
			       ;; use this for resetting on rcs-delta's revert
			       (setq hoc-index-buffer sim-index-buffer))
	  rcs-delta-posthook 'sim-set-known-file
	  kill-buffer-hook
	  (lambda () (and sim-index-buffer;; is there one?
			  (buffer-name sim-index-buffer);; still exists?
			  (set-ilv 'sim-emacs-buffers
				   (delete (current-buffer)
					   (get-ilv 'sim-emacs-buffers)))))
	  indent-tabs-mode nil
	  outline-regexp "\\(\\(//\\)\\|\\(:\\)\\)*[*]+"
	  comment-start "// "
	  comment-end ""
	  comment-start-skip "/+\\(\\*+ *\\)*"
	  comment-indent-function 'c-comment-indent
	  hoc-indent-enabled nil	; Turn indentation off so we can toggle
					; it on (& set up many other params).
	  require-final-newline t)	; Neuron (at least the older versions)
					; may do very strange things if its
					; input doesn't end in a newline.
    (hoc-toggle-indent)			; Turn indentation mode on.
    (if window-system
	(progn (sim-menu-bar (current-local-map))
	       (hoc-menu-bar (current-local-map))
	       (nrn-menu-bar (current-local-map))))
    (setq selective-display sel)
    (run-hooks 'hoc-mode-hook)))

(defun hoc-find-index ()
  "Returns appropriate index file, does not set-buffer."
  (let (indb indf bfl 
	     (cbuf (current-buffer))
	     (find (not current-prefix-arg)))
    ;; index file should be either in current directory or one above
    (if find  ;; look for index file automatically
	(cond 
	 (hoc-index-buffer (setq indf (buffer-file-name hoc-index-buffer)
				 hoc-index-buffer nil))
	 ((file-exists-p (setq indf (concat default-directory "index"))))
	 ((file-exists-p (setq indf (concat default-directory "../index"))))
	 ((progn
	    (setq bfl (buffer-list))
	    (while (and bfl;; search through the buffer list till none left
			(not (and (string-match
				   "index"
				   (or (buffer-file-name (car bfl)) ""))
				  (progn (set-buffer (car bfl))
					 (not buffer-read-only)))))
	      (setq bfl (cdr bfl)))
	    (setq indb (if bfl (car bfl) ()))
	    (if (not indb) ()
	      (setq indf (expand-file-name (buffer-file-name indb))))))
	 (t (setq indf nil))))
    (or indf  ;; if not index found then ask user
	(string-match "index" 
		      (setq indf (expand-file-name
				  (read-file-name 
				   "Index file (default none): "
				   (concat (getenv "HOME") "/") "" nil ""))))
	(setq indf nil))
    (if (not indf) (progn (set-buffer cbuf) ())
      ;; see if the index file already available in a buffer
      (cond
       ((setq indb (find-buffer-visiting indf)))
       ((setq indb (find-file-noselect indf))))
      
      (set-buffer indb)

      ;; see if the index file has already been initialized with all the stuff
      (if (assoc 'sim-index-buffer (buffer-local-variables))
	  (and (not (member cbuf sim-emacs-buffers))
	       (setq sim-emacs-buffers (nconc sim-emacs-buffers (list cbuf))))
	(nrn-find-index default-directory);; sets the local variables and everything else
	(mapcar 
	 (lambda (var) (make-local-variable (eval var)))
	 sim-localvar-list)
	(setq sim-default-directory default-directory
	      sim-emacs-buffers (list cbuf)
	      sim-index-buffer indb
	      sim-buffer nil sim-library-path nil sim-redirect-output nil))

      (set-buffer cbuf)
      indb)))

(defun hoc-toggle-indent ()
  "Toggles the function of the TAB key.  TAB can either insert a hardware tab
or calculate a line indentation as is done in c-mode."
  (interactive)
  (cond (hoc-indent-enabled		; Did it used to be on?
	 (setq hoc-indent-enabled nil	; Turn it off.
	       indent-line-function 'indent-to-left-margin)
	 (use-local-map hoc-indent-disabled-map)) ; Enable only C-x C-c.
	(t				; Used to be off:
	 (setq hoc-indent-enabled t	; Turn it on.
	       indent-line-function 'hoc-indent-line)
	 (use-local-map hoc-mode-map)))) ; Enable all the indentation keys.

(defun hoc-next-function ()
  "Finds the next hoc function by searching for a pattern matching
beginning of function."
  (interactive)
  (re-search-forward hoc-function-start-pattern))

(defun hoc-previous-function ()
  "Finds the prior hoc function by searching for a pattern matching
beginning of function."
  (interactive)
  (re-search-backward hoc-function-start-pattern))

(defun hoc-neuron-command ()
  "Send command to neuron"
  (interactive)
  (let ((comm))
    (setq sim-redirect-output 1
	  comm (completing-read "Command: " nrn-command-alist nil nil 
				(car hoc-command-hist) '(hoc-command-hist . 1)))
    (if (or (string= comm "") current-prefix-arg)(setq comm "run()"))
    (or (string= comm (car hoc-command-hist))
	(setq hoc-command-hist (cons  comm hoc-command-hist)))
    (hoc-sender (sim-proc) (concat comm "\n"))))
    
(defun hoc-toggle-send-all ()
  "Determine whether to send something just to associated buffer or to all
sim-buffers."
  (interactive)
  (setq hoc-send-all-sim-buffers (if hoc-send-all-sim-buffers nil t))
  (message "%s" (if hoc-send-all-sim-buffers "Send to ALL." "Send to ONE only.")))

(defun hoc-send-line ()
  "Similar to hoc-send-line-and-save but doesn't send to mult buffers
and will strip a leading // to uncomment a line"
  (interactive)
  (setq sim-redirect-output nil)
  (let (comm (eolptr (progn (end-of-line) (point)))
	     (cbuf (current-buffer)))
    (beginning-of-line)			; Move to first character.
    (setq comm (buffer-substring (point) eolptr))
    (if (string-match "^ *[/]+ *\\(.+\\)" comm)(setq comm (match-string 1 comm)))
    (if (string-match "^.+\\(//\\)" comm)(setq comm (substring comm 0 (match-beginning 1))))
    (set-buffer sim-buffer)
    (insert-before-markers comm "\n")
    (ring-insert comint-input-ring comm)
    (hoc-sender (sim-proc) (concat "{" comm "}"))
    (set-buffer cbuf)
    (forward-line 1)))

(defun hoc-send-line-and-save ()
  "Sends current line of text to the neuron interpreter and saves file."
  (interactive)
  (if (not (eq sim-redirect-output 7)) (setq sim-redirect-output nil))
  (save-excursion
    (let ((cbuf (current-buffer))
	  comm 
	  (eolptr (progn (end-of-line) (point)))) ; Mark end of line.
      (beginning-of-line)		; Move to first character.
      (setq comm (buffer-substring (point) eolptr))
      (save-buffer)
      (if hoc-send-all-sim-buffers
	  (mapcar (lambda (x) (set-buffer(process-buffer x))(insert-before-markers comm "\n"))
		  (sim-all-processes))
	(set-buffer sim-buffer)
	(insert-before-markers comm "\n"))
      (ring-insert comint-input-ring comm)
      (set-buffer cbuf)
      (hoc-sender (sim-proc) (concat comm))
      ))
  (or (eq sim-redirect-output 7) (forward-line 1)))

(defun hoc-sender (proc str)
  "Send PROCESS the contents of STRING as input.
Adapted from comint-send-string but only sends chunks that end in newlines.
Equivalent to `process-send-string', except that long input strings
are broken up into chunks of size `hoc-input-chunk-size'.  Processes
are given a chance to output between chunks.  This can help prevent processes
from hanging when you send them long inputs on some OS's."
  (let* ((len (length str))
	 (i 0)
	 (pproc (if hoc-send-all-sim-buffers (sim-all-processes) (list proc)))
	 (next-i (min len hoc-input-chunk-size)))
    (while (< i len)
      (if (string-match "\n" (substring str next-i))
	  (setq next-i (+ next-i (match-end 0)))
	(setq next-i len))
      (sit-for 0.1) ;; Linux sensitive; may need to readjust chunk size and/or sit-for
      (accept-process-output)
      (mapcar (lambda (x) (process-send-string x (substring str i next-i))) pproc) 
      (setq i next-i next-i (min len (+ i hoc-input-chunk-size))))
    (mapcar (lambda (x) (process-send-string x "\n")) pproc)))

(defun hoc-send-string (string-to-send)
  "Sends a string to the neuron process.
All tabs are removed from the string to prevent readline from doing variable
name expansion."
  (let* (beg-insert)
    (sim-switch-to-buffer t)
    (setq beg-insert (point))		; Remember where insertion started.
    (insert string-to-send)		; Put the text in.
    (untabify beg-insert (point))	; Remove tabs.
    (comint-send-input)))		; Send the input.

(defun hoc-send-region (beg end)
  "Sends a region or block of text to the hoc interpreter (neuron) directly.
If point=mark, a block will defined by a nonindented line and the
contents of the {} which follow."
  (interactive "r")
  (if (interactive-p) (save-buffer))
  (save-excursion
    (let ((sproc (sim-proc)) txt beg1 end1 pt)
      (setq sim-redirect-output 4 pt (point))
      (cond 
       ((= beg end);; look see if a:inside of a template or b:inside a block
	(and (setq beg (re-search-backward "^begintemplate\\s +\\([-_:a-z0-9A-Z]+\\)\\s *$" (point-min) t))
	     (setq txt (match-string 1))  ;; find beginning then end
	     (if (re-search-forward "^endtemplate\\s +\\([-_:a-z0-9A-Z]+\\)\\s *$" (point-max) t)
		 (or (string= txt (match-string 1))
		     (error "Template mismatch %s !~ %s" txt (match-string 1)))
	       (error "Template end not found for %s" txt))
	     (setq end (point)))
	(if (and beg (< beg pt) (< pt end)) ;; template found surrounding point?
	    (message "Sending `%s' template" txt)
	  (goto-char pt);; else look for block starting from point
	  (setq beg (re-search-backward "^{\\|^[a-z]+[^{]+{" (point-min) t))
	  (goto-char (1- (match-end 0)))
	  (forward-list) (setq end (point))
	  (if (and (< beg pt) (< pt end));; has block been found?
	      (message "SENDING BLOCK:: %s" (match-string 0))
	    (error "No block or template found"))))
       (t (message "SENDING:: %s >>>...<<< %s"
		   (buffer-substring beg (progn (goto-char beg)(end-of-line)(point)))
		   (buffer-substring 
		    (progn (goto-char (1- end))(beginning-of-line)(point)) (1- end)))))
      (setq txt (buffer-substring beg end))
      (hoc-sender sproc txt)
      (set-buffer (get-buffer-create "*nrn-output*")) (insert txt "\n")
      )))

(defun hoc-label-with-region (beg end)
"Removes newlines and excess spaces from region to send to use as a
label on the bottom of all of the graphs.  Also refreshes the date label."
  (interactive "r")
  (let ((str (buffer-substring beg end)) newstr)
    (while (string-match "[ \t]*\n[ \t]*" str)  ;; replace newlines with spaces
      (setq newstr (concat newstr " " (substring str 0 (match-beginning 0)))
	    str (substring str (match-end 0) (length str))))
    (setq newstr (concat newstr " " str))
    (hoc-sender
     (sim-proc)
     (concat "labels(\"" newstr "\")\n"))))

(defun hoc-send-proc ()
  "Sends a procedure or func to the interpreter (neuron).  Adds its
name onto nrn-command-alist"
  (interactive)
  (let (pt)
    (if (interactive-p) (save-buffer))
    (save-excursion
      (forward-line 1)
      (re-search-backward hoc-function-start-pattern)
      (nrn-update-command-alist (match-string 2))
      (setq hoc-command-hist (cons  (concat (match-string 2) "()") hoc-command-hist))
      (message "LOADING:: %s()" (buffer-substring (match-beginning 1) (match-end 2)))
      (forward-list 2) (forward-line 1) (beginning-of-line)
      (setq pt (point)))
    (setq sim-redirect-output 4)
    (hoc-sender (sim-proc) (buffer-substring (match-beginning 0) pt))
    (accept-process-output)
    ))

(defun hoc-rename-and-send-template ()
  "Sends a template - must be renamed so as not to cause trouble."
  (interactive)
  (let (pt e num name ch)
    (save-excursion
      (forward-line 1)
      (or (re-search-backward "^begintemplate +\\(.+\\)" nil t)
	  (error "begintemplate not found"))
      (setq name (match-string 1) ch (string-to-char (substring (match-string 1) -1))
	    newname (concat (substring name 0 -1) (char-to-string (1+ ch))))
      (message (format "%s -> %s" name newname))
      (replace-match (concat "begintemplate " newname))
      (re-search-forward (concat "^endtemplate +" name) nil t)
      (replace-match (concat "endtemplate " newname))
      (forward-line -1)
      (hoc-send-region (point)(point))
      (message newname))))

(defun hoc-send-buffer ()
"Sends the current buffer to neuron and switches you to neuron
buffer.  This saves the file before sending it. Neuron is actually
sent a command to read the file.

>> DANGER: if there is ANY error in the buffer NEURON will crash. <<
If uncertain, it is better to select the whole buffer and use
hoc-send-region, or else send procs individually"
  (interactive)
  (save-excursion
    (goto-char (point-max))		; Put a final newline at the end, if
    (beginning-of-line)			; there isn't one, or else hoc will
    (if (equal (point-max) (point))	; give a bizarre error message.
	nil
      (progn (goto-char (point-max)) (insert "\n"))))
  (save-buffer)				; Save the current buffer.
  (sim-open-file (buffer-file-name))
  (sim-switch-to-buffer t))

(defun hoc-xopen-region (beg end)
"Sends the region using an xopen()
>> DANGER: if there is ANY error in the buffer NEURON will crash. <<
"
  (interactive "r")
  (let ((fl (concat sim-default-directory "tmp.hoc")))
    (write-region beg end fl)
    (sim-open-file fl)))

;; Functions for indenting hoc lines:
(defun hoc-indent-exp ()
  "Indent each line of the current hoc grouping following point.  To indent a
whole block of code, point must be positioned just before the opening brace, as
in c-mode."
  (interactive)
  (mark-sexp 1)
  (indent-region (point) (mark) nil))

(defun hoc-indent-line ()
  "Indent current line as hoc code."
  (interactive)
  (let ((new-col nil))			; new-col is position we want to be at,
					; if different.
    (save-excursion
      (let* ((old-col (current-column))
	     (old-indent (current-indentation))
	     (indent (hoc-calculate-indent))
	     (beg-line (progn (beginning-of-line) (point))) ; Where line starts.
	     shift-amt)
	(if (not indent) nil		; Are we in a string?  If so, no change
	  (setq shift-amt (- old-indent indent)) ; Calculate difference.
	  (if (/= shift-amt 0)		; Any change?
	      (progn
		(skip-chars-forward " \t") ; Skip over any indentations.
		(delete-region beg-line (point)) ; Delete leading whitespace.
		(indent-to indent)	; Set correct indentation.
		;;
		;; If user was positioned in the whitespace preceding the first
		;; character, position him to the indent position.  Otherwise
		;; it is very confusing because the cursor doesn't end up
		;; where it is expected if the line is blank.
		;;
		(if (<= old-col old-indent)
		    (setq new-col indent)))))))
    (if new-col
	(move-to-column (max 0 new-col)))))
;;
;; Function to back up to last non-blank character not inside a comment.
;; Swiped from c-mode and simplified.
;;
(defun hoc-backward-to-noncomment (lim)
  (let (stop)
    (while (not stop)
  ;;
  ;; First look for old-style /* */ comments:
  ;;
      (while (not stop)
	(skip-chars-backward " \r\t\n\f" lim)
	(if (and (>= (point) (+ 2 lim))
		 (save-excursion
		   (forward-char -2)
		   (looking-at "\\*/")))
	    (while (and (search-backward "[^/]/*" lim 'move)
			(hoc-in-quote)))
	  (setq stop t)))
  ;;
  ;; Now make sure there isn't a // on the line:
  ;;
      (let ((opoint (point))		; Remember previous end of line.
	    (bol-point (progn (beginning-of-line) (point)))) ; Go to beginning.
	(while (and (search-forward "//" opoint t) ; Look for a //.
		    (if (hoc-in-quote) t ; Skip if quoted.
		      (forward-char -2)	; Found one.  Backup to first /.
		      (setq stop nil)))) ; Check again for /* */.
	(if stop			; Didn't find a //?
	    (goto-char opoint))))))	; Restore cursor position.
	  

(defun hoc-beginning-of-statement nil
  "Function to move point to the first non-blank character in the current
statement."
  (interactive)
  (while (progn
	   (beginning-of-line)		; See if current line could be the
					; first line of a statement.
	   (skip-chars-backward " \t\f\n"); Move to end of last line.
	   (= (preceding-char) ?\\)))	; Is this a continuation?
  (skip-chars-forward " \t\f\n"))	; Move to first non-blank char of
					; statement.
(defun hoc-in-old-style-comment ()
  "Returns the first character of a comment in the old style (/* */), if point
is somewhere in the body of a comment.  Returns nil otherwise."
  (let ((opoint (point))
	found-cmt)
    (while (and (setq found-cmt (search-backward "[^/]/*" (point-min) t))
		(hoc-in-quote)))	; Search backwards for opening comment.
    (if found-cmt			; Found beginning of comment?
	(setq found-cmt 
	      (if (search-forward "*/" opoint t) ; Loop for end of comment.
		  nil			; Found.  This comment's over.
		(point))))
    (goto-char opoint)			; Restore cursor position.
    found-cmt))				

(defun hoc-strip-prompts ()
  "Strips the prompts and returned values from a bunch of code
statements entered from the command line"
  (interactive)
  (save-excursion
    (let ((pt (point)))
      (while (re-search-forward "^[> 	]*oc>\\(.+\\)$" nil t)
	(delete-region pt (match-beginning 1))
	(setq pt (1+ (point))))
      (delete-region pt (point-max)))))

;;
;; The following function tries to determine if we are inside a double quoted
;; string or not.  The idea for this was borrowed somewhat from cplus-md.el.
;;
(defun hoc-in-quote ()
  "Returns true if point is in a quoted string, nil if not."
  (let ((opoint (point))		; Previous location of point.
	(n-quotes 0))
    (while (progn			; Back up to where string could
					; possibly begin.
	     (beginning-of-line)	; Move to beginning of current line.
	     (skip-chars-backward " \t\f\n\r") ; Find last non-blank char.
	     (= (preceding-char) ?\\)))	; Back up a line if last line ends in /
    (while (search-forward "\"" opoint t) ; Find the next quote.
      (if (/= (preceding-char) ?\\)	; Recognize a backslash escape.
	  (setq n-quotes (1+ n-quotes))) ; Remember that we saw another quote.
      (forward-char 1))			; Skip the matched character.
    (goto-char opoint)			; Restore point position.
    (not (zerop (logand n-quotes 1)))))	; We are in a quote if there are an
					; odd number of quotes.

(defun hoc-calculate-indent ()
  "Calculates indentation for current line.  Returns nil if we're inside a
string; always returns a number otherwise."
  (if (hoc-in-old-style-comment)
      (calculate-c-indent-within-comment)
    (save-excursion
      (beginning-of-line)
      (skip-chars-forward " \t")	; Move to first non-blank char.
      (let ((old-point (point)))
;;;
;;; The algorithm for indenting is a little simpler than C mode, because
;;; hoc doesn't have statements that go on for more than one line unless
;;; they're continued by a backslash.
;;;
	(if (= (following-char) ?\})	; Conclusion of block?
	    (progn
	      (forward-char 1)		; Move to the other side of {.
	      (backward-sexp)		; Go to the matching {.
	      (current-indentation))	; Use its indentation.
;;;
;;; This is not a closing brace.  If this isn't a continuation line, indent
;;; the same as the previous line unless there are unmatched opening braces.
;;;
	  (hoc-backward-to-noncomment (point-min)) ; Go to previous char.
	  (if (/= (preceding-char) ?\\)	; If it wasn't a backslash, just use
	      (let* ((eol-prec (point))	; Remember where this line ends.
		     (n-unmatched-open 0) ; Number of unmatched open braces.
		     current-char
		     (beg-prec (progn (hoc-beginning-of-statement) (point))))
					; Go to beginning of prior statement.
		(while (re-search-forward "[{}]" eol-prec t)
					; Find a significant character.
		  (if (hoc-in-quote) nil ; Ignore if in a quoted string.
		    (setq current-char (preceding-char))
		    (cond ((= current-char ?\{) ; Opening brace?
			   (setq n-unmatched-open (1+ n-unmatched-open)))
			  ((= current-char ?\}) ; Closing brace?
			   (setq n-unmatched-open
				 (max 0 (1- n-unmatched-open)))))))
		(goto-char beg-prec)	; Go back to beginning of prior stmt.
		(+ (current-indentation) (* hoc-block-indent n-unmatched-open))
	      )
;;;
;;; This is a continued statement.  Find the last unbalanced parenthesis and
;;; use its indentation, or else just indent from the beginning of the block.
;;;
	    (hoc-beginning-of-statement)
	    (let* ((statement-begin-column (current-column))
		   (parse-sexp-ignore-comments t)
		   (sexp-list (parse-partial-sexp (point) old-point))
		   (inner-sexp (car (cdr sexp-list))))
	      (cond ((nth 3 sexp-list)	; Are we in a string?
		     nil)		; Don't change indentation.
		    (inner-sexp		; Are we inside a parenthesis?
		     (goto-char inner-sexp) ; Move to open parenthesis.
		     (+ (current-column) 1)) ; Indentation level is 1 more.
		    (t			; Otherwise, just indent a little
		     (goto-char old-point) ; Move cursor to first nonblank char
					; of line we're indenting.
		     (if (looking-at "else")
			 statement-begin-column
		       (+ statement-begin-column hoc-continued-statement-offset)))
	      )
	    )
	  )
	)
      )
    )
  )
)

(provide 'hoc)
