;; $Id: nrn.el,v 1.174 2002/03/29 18:04:12 billl Exp $

;; This file is part of GNU Emacs.

;; GNU Emacs is distributed in the hope that it will be useful,
;; but WITHOUT ANY WARRANTY.  No author or distributor
;; accepts responsibility to anyone for the consequences of using it
;; or for whether it serves any particular purpose or works at all,
;; unless he says so in writing.  Refer to the GNU Emacs General Public
;; License for full details.

;; Everyone is granted permission to copy, modify and redistribute
;; GNU Emacs, but only under the conditions described in the
;; GNU Emacs General Public License.   A copy of this license is
;; supposed to have been given to you along with GNU Emacs so you
;; can know your rights and responsibilities.  It should be in a
;; file named COPYING.  Among other things, the copyright notice
;; and this notice must be preserved on all copies.
;; 
;; simctrl package by Bill Lytton - 11/1/93
;;

(require 'comint)
(require 'nrnrcs)
(require 'simctrl)
(require 'nrnutil)
(require 'hoc)
(load "nrnlocal" t t nil)

;;; INFERIOR NEURON MODE STUFF
;;;============================================================================

(defvar inferior-nrn-mode-hook nil
  "Hook for customising inferior-nrn mode.")
(defvar inferior-nrn-mode-map nil)
(defvar inferior-nrn-filter-regexp "\\`\\s *\\S ?\\S ?\\s *\\'"
  "Input matching this regexp are not saved on the history list.
Defaults to a regexp ignoring all inputs of 0, 1, or 2 letters.")
(defvar nrn-program-name nil
  "Program invoked by the run-neuron command")
(defvar nrn-same-prog nil
  "*Set to true to ensure running same nrn-program-name as last time")
(defvar nrn-nload nil
  "*Set to true to not load initialization sequence for debugging")
(defvar nrn-remote-machine nil
  "Remote machine running neuron, available in sim-buffer only")
(defvar nrn-uname nil "Picked up from ENV variable $CPU")
(defvar nrn-osname nil "Picked up with uname -s")
(defvar nrn-command-alist ()
  "List of commands known to neuron.  This is constructed by looking
at source files (src/nrnoc/hocusr.h, src/oc/hoc_init.c) as well as
extracting procs from nrn-init-file and from the files listed in
sim-files-alist.  New procs loaded using sim-load-proc will also be
added to this list.  List used by nrn-complete-symbol")
(defvar nrn-neuronhome (concat (getenv "NEURONHOME") "/")
  "*Directory containing source code for NEURON.  Ancillary programs
and files are expected to be in nrn-neuronhome/lib/hoc and 
nrn-neuronhome/../simctrl/hoc")
(defvar nrn-src "/usr/site/nrniv/nrn-4.3.1/src/" "*neuron src files")
(defvar nrn-simctrl (concat (getenv "SIMCTRL") "/")
  "*Directory containing source code for SIMCTRL.")
(defvar nrn-init-file (concat nrn-simctrl "hoc/nrnoc.hoc")
  "*Initialization file for NEURON.")
(defvar nrn-last-input "" "The last input sent to neuron.")
(defvar nrn-assignment t
  "A string being assigned a value in neuron, previously used to have
NEURON make alterations in params.hoc file. Currently simply issue warning.")
(defvar nrn-filter-prog ""
  "For storing programs that are created on the fly.  A less crashable
alternative to creating an external file with the prog and doing an xopen")
;; What numbers look like in NEURON.
(defvar nrn-init-flags "" "*Flags to use when calling special.")
(setq nrn-num-regexp "[-.]*[0-9]+[.]*[-+e0-9]*")

(defun nrn-mode-commands (map)
  (define-key map "\e\C-q"   'indent-c-exp)
  (define-key map "\177"     'backward-delete-char-untabify)
  (define-key map "\C-xl"    'sim-load-sim)
  (define-key map "\C-xd"    'sim-direct-output)
  (define-key map "\C-x\C-l" 'sim-load-templates)
  (define-key map "\C-xf"    'sim-load-file)
  (define-key map "\C-x\C-a" 'nrn-append-alist)
  (define-key map "\C-xs"    'sim-save-sim)
  (define-key map "\C-xz"    'sim-switch-to-buffer)
  (define-key map "\C-xc"    'sim-current-files)
  (define-key map "\C-x "    'sim-rcs-status)
  (define-key map "\C-x?"    'sim-template-status)
  (define-key map "\t"       'nrn-complete-symbol)
  (define-key map "\e\t"     'nrn-complete-symbol))

;; interface key bindings that differ from comint standards
(defun nrn-inter-map (mp) 
  (define-key mp "\C-cp" 'previous-prompt)
  (define-key mp "\C-cn" 'next-prompt)
  (define-key mp "\C-c\C-@" 'process-region)
  (define-key mp "\C-c\C-b" 'user-interrupt)
  (define-key mp "\C-x\C-l" 'mv-output)
  (define-key mp "\C-c\C-l" 'mv-output)
  (define-key mp "\C-c\C-d" 'send-eof)
  (define-key mp "\C-d" 'delete-char))

(defun nrn-menu-bar (map)
  (define-key map [menu-bar nrn] (cons "Nrn" (make-sparse-keymap "nrn")))
  (define-key map [menu-bar nrn dout] 
	'("Output to sim buffer" . sim-direct-output))
  (define-key map [menu-bar nrn grapher]
    '("Grapher..." .
      (lambda()(interactive)(comint-send-string(sim-proc)"makegrapher()\n"))))
  (define-key map [menu-bar nrn Expr]
    '("Expr Plot..." .
      (lambda()(interactive)(comint-send-string(sim-proc)"newgraph(1)\n"))))
  (define-key map [menu-bar nrn phs]
    '("Phase plane" .
      (lambda()(interactive)(comint-send-string(sim-proc)"newphaseplane()\n"))))
  (define-key map [menu-bar nrn vec]
    '("Vector Movie" .
      (lambda()(interactive)(comint-send-string(sim-proc)"newvectorplot()\n"))))
  (define-key map [menu-bar nrn Shape]
    '("Shapeplot" .
      (lambda()(interactive)(comint-send-string(sim-proc)"newshapeplot()\n"))))
  (define-key map [menu-bar nrn State]
    '("Stateaxis" .
      (lambda()(interactive)(comint-send-string(sim-proc)"newPlotS()\n"))))
  (define-key map [menu-bar nrn Current]
    '("Currentaxis" .
      (lambda()(interactive)(comint-send-string(sim-proc)"newPlotI()\n"))))
  (define-key map [menu-bar nrn Voltage]
    '("Voltageaxis" .
      (lambda()(interactive)(comint-send-string(sim-proc)"newPlotV()\n"))))
  (define-key map [menu-bar nrn lgrs]
    '("Label All w/Runnum" .
      (lambda()(interactive)(comint-send-string(sim-proc)"labelgrs()\n"))))
  (define-key map [menu-bar nrn lwith]
    '("Label With..." .
      (lambda()(interactive)(comint-send-string(sim-proc)"labelwith()\n"))))
  (define-key map [menu-bar nrn lpanel] 
    '("Label Panel" .
      (lambda()(interactive)(comint-send-string(sim-proc)"labelpanel()\n"))))
  (define-key map [menu-bar nrn tchange]
    '("Tstop Changed" .
      (lambda()(interactive)(comint-send-string(sim-proc)"tstop_changed()\n"))))
  (define-key map [menu-bar nrn comm]
    '("Run Commands" .
      (lambda()(interactive)(comint-send-string(sim-proc)"stoppanel()\n"))))
  (define-key map [menu-bar nrn spanel] 
    '("Simulation Main Menu" .  
      (lambda()(interactive)(comint-send-string(sim-proc)"sim_panel()\n"))))
)

(defun nrn-mode-variables ()
  ;; (setq local-abbrev-table nrn-mode-abbrev-table)
  (abbrev-mode 1)
  (make-local-variable 'nrn-program-name)
  (make-local-variable 'nrn-init-file)
  (make-local-variable 'nrn-last-input)  
  (make-local-variable 'sim-default-element)
  (make-local-variable 'comint-process-echoes)
  (setq comint-process-echoes nil)
  (setq sim-redirect-output 0)
)

(cond ((not inferior-nrn-mode-map)
       (setq inferior-nrn-mode-map
	     (copy-keymap comint-mode-map))
       (nrn-inter-map inferior-nrn-mode-map)
       (nrn-mode-commands inferior-nrn-mode-map))) 

(defun inferior-nrn-mode ()
  "Major mode for interacting with an inferior Neuron process.
The following commands are available:
\\{inferior-nrn-mode-map}
A NEURON process can be fired up with M-x run-neuron.
Customisation: Entry to this mode runs the hooks on comint-mode-hook and
inferior-nrn-mode-hook (in that order).
Commands:
Return after the end of the process' output sends the text from the 
    end of process to point.
Delete converts tabs to spaces as it moves back.
Tab or Esc-Tab does completion for neuron commands
    This listing may not be complete.
    To see a list of all commands known to emacs hit tab from
    beginning of line.
If you accidentally suspend your process, use \\[comint-continue-subjob]
to continue it."

  (interactive)
  ;; Customise in inferior-nrn-mode-hook
  (setq comint-prompt-regexp "^[a-z0-9]+> *") 
  (nrn-mode-variables)
  (use-local-map inferior-nrn-mode-map)
  (setq major-mode 'inferior-nrn-mode
	mode-name "Inferior Neuron"
	mode-line-process '(": %s")
	comint-input-filter (function nrn-input-filter)
	comint-input-sentinel (function nrn-input-sentinel)
	comint-input-sender (function nrn-input-sender)
	comint-get-old-input (function comint-get-old-input-default))
  (set-marker (process-mark (get-buffer-process (current-buffer))) (point))
  (set-process-filter (get-buffer-process (current-buffer)) 'nrn-filter)
  (run-hooks 'inferior-nrn-mode-hook))

(defun nrn-input-sentinel (str)
  (cond 
    ((string-match "access +\\([][a-z_0-9\\.]+\\)" str) 
     (setq sim-default-element (substring str (match-beginning 1) (match-end 1)))
     (message "sim-default-element: %s" sim-default-element))
    ((or
     (and nrn-assignment (string-match "[^n]insert" str))
     (and nrn-assignment
	  (not (string-match "\\(for\\)\\|\\(new\\)" str))
	  (string-match "[^!><=]=[^=][ \t]*[0-9]" str) ;; an assignment
	  (string-match "\\([][a-z_0-9\\.]+\\)[ \t]*=" str)))
     (ding) (message "WARNING: set in a file, not directly in NEURON"))
))

(defun nrn-input-filter (str)
  "Don't save anything matching inferior-nrn-filter-regexp"
  (not (string-match inferior-nrn-filter-regexp str)))

(defun nrn-input-sender (proc string)
  (setq nrn-last-input (concat string "\n"))
  (comint-send-string proc (concat string "\n")))

;; modified comint-filter so can throw out a string or send it elsewhere
(defun nrn-filter (process string)
  (if (file-exists-p "nrn.log")(write-region string nil "nrn.log" t 0))
  (cond 
   ;; pattern *&* signals a command for emacs requested from the nrn buffer
   ((string-match "\\*&\\* \\([-a-z]+\\)" string)
    (setq string (intern (substring string (match-beginning 1)(match-end 1))))
    (if (commandp string) (call-interactively string nil)
      (message "Error in menu call for %s" string)))
   ;; pattern !&! is a line to be displayed in the message field of emacs
   ((string-match "!&! *\\(.+\\)" string)
    (message "%s" (match-string 1 string)))
   ;; pattern ^&^ is an elisp command
   ((string-match "\\^&\\^ *\\(.+\\)" string)
    (set-buffer (process-buffer process))  ;; move to process buffer
    (set-buffer sim-index-buffer)          ;; move to index file
    (message string)
    (eval (read (match-string 1 string)))) ;; evaluate string in this context
   ;; pattern &INDEX& is a line to put in the index file
   ((string-match "&INDEX& *\\([^&]+\\)&" string) 
    (set-buffer (process-buffer process))  ;; move to process buffer
    (set-buffer sim-index-buffer)          ;; move to index file
    (clear-visited-file-modtime)
    (write-region (concat (match-string 1 string) "\n")
		  nil buffer-file-name t 0)
    (setq string 
	  (concat (substring string 0 (match-beginning 0))(substring string (match-end 0) (length string))))
    (comint-output-filter process string)
    (sim-index-revert))
   ;; '*!!* BEGIN' signals beginning of strings to be regurgitated back into NEURON
   ;; '*!!* END' signals end - must be inside a procedure
   ;; proc test () {                  
   ;;   print "!{!\n"           
   ;;   print "print \"hello\"\n" 
   ;;   print"!}!\n" }            
   ((string-match "!\\([{}]\\)!" string)
    (cond
     ((string= "{" (substring string (match-beginning 1)(match-end 1)))
      (message (substring string (- (match-end 0))))
      (setq nrn-filter-prog (substring string (match-end 0))
	    sim-redirect-output 6))
      ((string= "}" (substring string (match-beginning 1)(match-end 1)))
      (setq sim-redirect-output nil
	    nrn-filter-prog (concat nrn-filter-prog
				    (substring string 0 (match-beginning 0))))
      (comint-send-string (sim-proc) nrn-filter-prog))
     (t (error "!{}!"))))
   ;; sim-redirect-output flag
   ;; 0: initialization - gets rid of 'first instance'
   ;; 1: redirect output to buffer *nrn-output*
   ;;    tries to discard excess prompts
   ;; 2: display output in minibuffer 
   ;; 3: discard output - don't show anywhere
   ;; 4: used for loading hoc code, looks for "error"
   ;; 5: used to pick up the value of a variable from the 
   ;;    simulator
   ;; 6: add lines to nrn-filter-prog
   ;; default: comint-filter, just put the text there
   ((eq sim-redirect-output 0) ;; **** 0 **** 
    (cond ((string-match "^Init" string)
	   (setq sim-redirect-output nil))
	  ((or (string-match "first instance" string)
	       (string-match "^\\(oc>\\)+ *$" string)
	       (string-match "^[ \t]*1 *$" string)
	       (string-match "Getting" string))())
	  (t (comint-output-filter process string))))
   ((eq sim-redirect-output 1) ;; **** 1 ****
    (cond ;; redirect output to another buffer
     ((string= nrn-last-input string)
      (set-buffer (get-buffer-create "*nrn-output*")) (erase-buffer))
     ((string-match "\\(.*\\)\\(\nneuron[0-9]+ > *\\)" string) 
      (pop-to-buffer (get-buffer-create "*nrn-output*"))
      (goto-char (point-max))
      (insert "\n")
      (insert (substring string 0 (match-end 1)))
      (goto-char (point-min))
      (pop-to-buffer sim-buffer)
      (insert-before-markers 
       (substring string (match-beginning 2) (match-end 2))))
     (t
      (set-buffer (get-buffer-create "*nrn-output*"))
      (goto-char (point-max))
      (insert string)
      (goto-char (point-max))))
    (setq sim-redirect-output nil))
   ((eq sim-redirect-output 2)  ;; **** 2 ****
    (cond  ;; put output in message area
     ((string-match "[ \t]*\\(.*\\)\n*oc>" string)
      (message (substring string (match-beginning 1) (match-end 1)))
      (setq sim-redirect-output nil))
     (t (message string))))
   ((eq sim-redirect-output 3) ()) ;; **** 3 **** discard output
   ((eq sim-redirect-output 4)   ;; **** 4 ****
    (sit-for 0)
    (if (string-match "\\(\\([\t >]*\\(oc\\)*>+[\t >]*\\)+\\)" string)
	(setq string (substring string (match-end 0))))
    (if (string= string "") ()
      (let ((buf (current-buffer)) nbuf)
	(pop-to-buffer (get-buffer-create "*nrn-output*"))
	(setq sim-redirect-output 4);; set it in this new buffer
	(goto-char (point-max))
	(insert string)
	(goto-char (point-max))
	(pop-to-buffer buf)))
    (if (string-match "error" string)
	(message "ERROR: see buffer *nrn-output*"))
    )
   ((eq sim-redirect-output 5) ;; **** 5 ****
    (cond ((string-match "END*&*" string) 
	   (setq sim-redirect-output nil)
	   (apply nrn-set-var (list string)))))
   ((eq sim-redirect-output 6) ;; **** 6 ****
    (setq nrn-filter-prog (concat nrn-filter-prog string)))
   ((eq sim-redirect-output 7) ;; **** 7 ****
    (end-of-line)
    (narrow-to-region (point)(point))
    (insert " // " string)
    (white-to-space)
    (replace-regexp-loop "// \\s *oc>\\s *" "")
    (widen)
    (forward-char 1))
   ;; check for current input else put it out
   (t
    (if (string= nrn-last-input string) () ;; discard
      (comint-output-filter process string)))
))

(defun nrn-args-to-list (string)
  (let ((where (string-match "[ \t]" string)))
    (cond ((null where) (list string))
	  ((not (= where 0))
	   (cons (substring string 0 where)
		 (nrn-args-to-list (substring string (+ 1 where)
						 (length string)))))
	  (t (let ((pos (string-match "[^ \t]" string)))
	       (if (null pos)
		   nil
		 (nrn-args-to-list (substring string pos
						 (length string)))))))))

(defun run-remote-neuron ()
  "Run neuron on a remote machine.  Performs an rlogin - requires that
permission be previously granted or that can write .rhosts
To do the latter, requires common NFS mounts.
WARNING: will overwrite .rhosts and then remove it."
  (interactive)
  (let* ((mname (completing-read "Host: " hostlist)))
    (run-neuron mname)))

(defun run-neuron (&optional mname)
  "Run an inferior Neuron process, input and output via buffer neuron.
  With argument, allows you to set the default dir.  If called from
  a file associated with a simulation (index, hoc or mod file) will
  automatically set default dir from that file.
  Options can be set with prefix arg:
  ^U : ask for simulation directory (will use default name neuron if available)
  ^U^U : ask for simulation name (will take dir from current buffer)
  ^U^U^U : run regular nrniv and don't load stuff"

  (interactive)

  (let (indbuf simbuf simdir)

    ;; take input from user for defaults if requested with prefix arg
    (setq simdir (nrn-read-defaults current-prefix-arg)
	  nrn-noload nil)
    ;; find index file, set sim-index-buffer
    (setq indbuf (nrn-find-index simdir))
    ;; startup special, set sim-buffer
    (if mname
	(setq simbuf (nrn-remote-start mname simdir current-prefix-arg))
      (setq simbuf (nrn-start-special simdir current-prefix-arg)))
    (setq sim-index-buffer indbuf);; sets the local one in sim-buffer

    (set-buffer indbuf)
    ;; reset name to reflect sim name
    (rename-buffer  (concat (buffer-name simbuf) "-index")) 
    (message "Using %sindex in buffer %s" 
	     simdir (buffer-name sim-index-buffer))
    ;; copy all local vars from sim-buffer to index and other files
    (sim-copy-local-variables sim-localvar-list simbuf indbuf) 
    (mapcar (lambda (fl) 
	      (and (buffer-name fl);; skipped any that have been killed
		   (sim-copy-local-variables 
		    sim-localvar-list simbuf fl)))
	    sim-emacs-buffers)
					
    ;; show stuff to user, load init file
    (nrn-show-process-defaults simbuf)
    (if mname ()            ;; initial-filter already active
      (setq telnet-count 3) ;; start where neuron already running
      (set-process-filter (get-buffer-process simbuf)'nrn-initial-filter))
    ))

;; if run neuron is called with a prefix will ask the user for init
;; dir and init file 
(defun nrn-read-defaults (arg)
  (let ((sdefdir
	 (cond 
	  ((eq (car arg) 4) (expand-file-name (read-file-name 
				  "Default dir: " 
				  sim-default-directory sim-default-directory t)))
	  ((assoc 'sim-default-directory (buffer-local-variables))
	   sim-default-directory)
 	  ((string-match "index" (buffer-name))
 	   (expand-file-name default-directory))
	  ((getenv "NEURONMAIN"))
	  (t (error "Unable to determine default-directory (use ^U prefix key).")))))
    (and (getenv "NEURONHOME") 
	 (setq nrn-neuronhome (concat (getenv "NEURONHOME") "/")))
    (cond
     ((not (file-exists-p nrn-init-file))
      (error "Can't find init %s." nrn-init-file))
     ((not (file-directory-p sdefdir))
      (error "Can't find sim-dir: %s." sdefdir)))
    sdefdir))

(defun nrn-find-index (ddir)
  (save-excursion
    (let ((ibuf
	   (or (find-buffer-visiting (concat ddir "index"))
	       (find-file-noselect (concat ddir "index")))))
      (set-buffer ibuf)

      ;; look for another process that goes with this index file
      (if (cdr (assoc 'sim-buffer (buffer-local-variables)))
	  (let ((oldsbuf (cdr (assoc 'sim-buffer (buffer-local-variables)))))
	    (if (buffer-name oldsbuf);; the buffer still exists
		(if (comint-check-proc oldsbuf);; and something's running in it
		    (error "process in '%s' buffer using index buffer '%s'";; signal error
			   (buffer-name oldsbuf) sim-index-buffer)
		  (kill-buffer oldsbuf));; else nothings running there anyway
	      )))

      (set-buffer ibuf)

      (if (assoc 'sim-buffer (buffer-local-variables)) ();; previously initialized?
	(kill-all-local-variables);; clean slate
	;; don't let them kill the index buffer
	(make-local-variable 'kill-buffer-query-functions)
	(setq kill-buffer-query-functions 
	      '((lambda () 
		  (read-string "THIS WILL KILL ENTIRE SIMULATION!! (^G TO ABORT)"))))
	(make-local-variable 'kill-buffer-hook)
	(setq kill-buffer-hook 
	      '((lambda () 
		  (and sim-buffer (get-buffer sim-buffer)(kill-buffer sim-buffer))
		  (while sim-emacs-buffers
		    (and (not (buffer-modified-p (car sim-emacs-buffers)))
			 (kill-buffer (car sim-emacs-buffers)))
		    (setq sim-emacs-buffers (cdr sim-emacs-buffers))))))
	(setq after-save-hook nil)
	(make-local-variable 'after-save-hook)
	(setq after-save-hook 
	      '((lambda () (sim-revert-buffers buffer-file-name))))
	(setq sim-index-buffer ibuf)
	(nrn-index-keys);; sets key bindings
	(mapcar 
	 (lambda (var) (make-local-variable (eval var)))
	 sim-indexvar-list))
      (setq sim-redirect-output nil  ;; always start this way
	    sim-template-loaded-flag nil)
      ;; pick up the current simulation files
      (sim-read-recent-versions)
      (goto-char (point-max))
      (sim-find-template-vers)
      ibuf)))

(defun nrn-start-special (sdefdir arg)
  ;; find the "special" (the executable for NEURON)
  (let* ((slibpath
	  (if (getenv "HOC_LIBRARY_PATH")
	      (string-to-listr " " (getenv "HOC_LIBRARY_PATH"))
	    (append 
	     (if (file-exists-p (concat sdefdir "hoc"))
		 (list (expand-file-name (concat sdefdir "hoc"))))
	     (if (file-exists-p (concat nrn-neuronhome "../local/hoc"))
		 (list (concat nrn-neuronhome "../local/hoc")))
	     (list (concat nrn-neuronhome "../simctrl/hoc")
		   (concat nrn-neuronhome "hoc")))))
	 (process-environment
	  (nconc
	   ;; handle neuron specific environmental variables
	   (list (format "NEURONMAIN=%s" sdefdir)
		 (format "NEURON=%s" nrn-neuronhome))
	   (if (getenv "HOC_LIBRARY_PATH") ()
	     (list (format "HOC_LIBRARY_PATH=%s"
			   (list-to-string slibpath))))
	   ;; taken from comint.el: comint-exec-1
	   (if (and (boundp 'system-uses-terminfo) system-uses-terminfo)
	       (list "EMACS=t" "TERM=unknown"
		     (format "COLUMNS=%d" (frame-width)))
	     (list "EMACS=t" "TERM=emacs"
		   (format "TERMCAP=emacs:co#%d:tc=unknown" (frame-width))))
	   process-environment))
	 sbuf)
    (setq sbuf (nrn-sim-name arg))
    (nrn-which-special sdefdir (getenv "CPU") (getenv "OSTYPE"))
    (set-buffer (get-buffer-create sbuf));; sbuf starts out as a name...
    ;; launch Neuron process
    (setq default-directory sdefdir)
    (if (> (length nrn-init-flags) 0)
	(start-process sbuf sbuf nrn-program-name nrn-init-flags)
      (start-process sbuf sbuf nrn-program-name))
    (setq sbuf (current-buffer));; ...but is then used as a buffer
    (comint-mode)
    (inferior-nrn-mode)
    (nrn-init-buffer sdefdir sbuf slibpath)
    sbuf
    ))

(defun nrn-set-special (spec)
  (interactive "fSpecial: ")
  (or (file-executable-p spec)(error "Must be an executable special."))
  (setq-default nrn-program-name (expand-file-name spec)))

(defun nrn-which-special (sdefdir uname osname)
  "Used to find special and also do other tasks that are uname dependent."
  (setq nrn-uname uname)
  (setq nrn-osname osname)
  (cond
   ((string= osname "Linux")
    (setq hoc-input-chunk-size 256)))
  ;; find the special
  (cond
   ((and current-prefix-arg (= 64 (car current-prefix-arg)))
    (setq nrn-noload t
	  nrn-program-name (concat (getenv "NB") "/nrniv")))
   ((and current-prefix-arg (= 16 (car current-prefix-arg)))
    (setq nrn-noload t
	  nrn-program-name 
	  (read-file-name "Neuron special: " (getenv "NB") "special" t))) 
   ((and nrn-same-prog nrn-program-name 
	 (file-regular-p nrn-program-name)
	 ;; don't use if have changed dirs
	 (string= (expand-file-name (concat sim-default-directory "mod/"))
		  (file-name-directory nrn-program-name))) nrn-program-name)
   ((file-exists-p (setq nrn-program-name ;; look above mod for 'special'
			 (concat sdefdir "special"))))
   ;; changed location from nrn4 to nrn5:
   ((file-exists-p (setq nrn-program-name ;; new location of specials
			 (concat sdefdir "mod/" nrn-uname "/special"))))
   ((file-exists-p (setq nrn-program-name ;; new location of specials
			 (concat sdefdir "mod/" nrn-uname "/.libs/special"))))
   ((error "mod/special doesn't exist, change sim-default-directory (%s)" 
	   sdefdir))))

(defun nrn-sim-name (arg)
  ;; determine the name for the sim buffer
  (let (sbuf
	(choice1 
	 (cond ((or (get-buffer "neuron-index") 
		    (and buffer-file-name
			 (file-symlink-p buffer-file-name)))
		(file-name-nondirectory 
		 (substring (file-name-directory buffer-file-name) 0 -1)))
	       ((string-match "\\([^-]+\\)-index" (buffer-name sim-index-buffer))
		(match-string 1 (buffer-name sim-index-buffer)))
	       ((string-match "index<\\([^>]+\\)>" (buffer-name sim-index-buffer))
		(match-string 1 (buffer-name sim-index-buffer)))
	       (t "neuron"))))
    (if (comint-check-proc choice1)
	(while (get-buffer 
		(setq sbuf (read-string "Sim buffer name?: " 
					(generate-new-buffer-name choice1)))) t)
      (if (get-buffer choice1) (kill-buffer choice1))
      (setq sbuf choice1))
    sbuf
    ))

(defun nrn-init-buffer (sdefdir sbuf slibpath)
  ;; make the local variables and set them
  (mapcar 
   (lambda (var) (make-local-variable (eval var)))
   sim-localvar-list)
  (setq default-directory sdefdir
	sim-default-directory sdefdir
	sim-buffer sbuf
	sim-redirect-output 0
	sim-library-path slibpath
	sim-file-open "xopen(\"##\")\n")

  ;; set up things used by sim including key bindings and menu
  (and window-system
       (sim-menu-bar inferior-nrn-mode-map)
       (nrn-menu-bar inferior-nrn-mode-map)))


(defun nrn-remote-start (mname sdefdir arg)
  "Log onto remote machine and exec 'special' (the executable for NEURON)
Requires NFS: will overwrite .rhosts."
  (let* ((slibpath
	  (if (getenv "HOC_LIBRARY_PATH")
	      (string-to-listr " " (getenv "HOC_LIBRARY_PATH"))
	    (append 
	     (if (file-exists-p (concat sdefdir "hoc"))
		 (list (expand-file-name (concat sdefdir "hoc"))))
	     (if (file-exists-p (concat nrn-neuronhome "../local/hoc"))
		 (list (concat nrn-neuronhome "../local/hoc")))
	     (list (concat nrn-neuronhome "../simctrl/hoc")
		   (concat nrn-neuronhome "../local/hoc")
			   (concat nrn-neuronhome "hoc")))))
	 sbuf)
    ;; will be used by nrn-initial-filter
    (setq 
     nrn-remote-env ;; Environment to place on remote machine.
     (nconc
      ;; handle neuron specific environmental variables
      (list (format "setenv PWD %s\n" sdefdir)
	    (format "setenv NEURON %s\n" nrn-neuronhome)
	    (format "cd %s\n" sdefdir);; just setting PWD is not enough
	    (format "setenv TERM unknown\n")
	    (format "setenv EMACS=t\n")
	    (format "setenv TERMCAP=emacs:co#%d:tc=unknown\n" (frame-width)))
      (if (getenv "HOC_LIBRARY_PATH") ()
	(list (format "setenv HOC_LIBRARY_PATH \"%s\"\n"
		      (list-to-string slibpath))))
	  ))
    (setq sbuf (nrn-sim-name arg))
    (call-process "xhost" nil nil nil mname)
    (switch-to-buffer (make-comint sbuf "ssh" () mname))
    ;; (sleep-for 1.2)
    (rename-buffer sbuf) ;; sbuf starts out as a name...
    (setq sbuf (current-buffer) ;; and becomes a buffer
	  nrn-program-name sdefdir  ;; temporary storage so filter can pass it
	  telnet-count 0)
    (nrn-init-buffer sdefdir sbuf slibpath)
    (set-process-filter (get-buffer-process sbuf) 'nrn-initial-filter)
    (make-local-variable 'nrn-remote-machine)
    (make-local-variable 'nrn-uname)
    (make-local-variable 'nrn-osname)
    (setq nrn-remote-machine mname)
    sbuf
    ))

(defun nrn-initial-filter (proc string)
  ;; use cond to do things in steps controlled with telnet-count
  ;; by checking and then incrementing telnet-count don't get multiple 
  ;; entry into same condition
  (set-buffer (process-buffer proc))
  (cond
   ;; 0: QUERY UNAME, SETENV DISPLAY AND OTHERS
   ((= telnet-count 0)
    (setq telnet-count 1)
    (comint-send-string proc "echo uname $CPU,,$OSTYPE\n") ;; get the machine
    (comint-send-string proc "stty nl -echo\n")
    (mapcar (lambda (x) (comint-send-string proc x)) nrn-remote-env))

   ;; 1: FIND NRN-PROGRAM-NAME
   ((string-match "uname \\([a-z0-9_]+\\),,\\([a-z0-9_]+\\)" string)
    (setq telnet-count 2)
    ;; nrn-program-name used for temp storage for sim-dir
    (nrn-which-special nrn-program-name (match-string 1 string)(match-string 2 string))
    (insert-before-markers "\nUsing " nrn-program-name "\n")
    (comint-send-string proc "\n"))

   ;; 2: EXEC THE NEURON PROGRAM
   ((and (string-match shell-prompt-pattern string)(= telnet-count 2))
    (setq telnet-count 3)
    (comint-send-string proc (concat "exec " nrn-program-name "\n")))

   ;; 3: FOLLOWING INITIALIZATION START NRNOC.HOC
   ((= telnet-count 3)
    (if nrn-noload ()
      (nrn-def-sim)
      (sim-open-file nrn-init-file)
      (if (file-exists-p "init.hoc") (sim-open-file "init.hoc") (message "NO init.hoc")))
    (inferior-nrn-mode)
    (if nrn-noload (setq sim-redirect-output nil)))
   ;; when nothing else to do insert string
   (t (insert string))))

(defun nrn-def-sim ()
  "Define things known to emacs in the simulation buffer."
  (interactive)
  (comint-send-string (sim-proc) 
		      (format "\n
strdef simname, output_file, datestr, uname, osname, comment\n
simname = \"%s\"
runnum = %d
datestr = \"%s\"
output_file = \"%s\"\n
uname = \"%s\"\n
osname = \"%s\"\n
xwindows = %d\n"
			      (buffer-name sim-buffer)
			      (get-ilv 'sim-runnum)
			      (date-id-string)
			      (get-ilv 'sim-output-file)
			      nrn-uname
			      nrn-osname
			      (if window-system 1 0))))

(defun nrn-show-process-defaults (simbuf)
  ;; tell the user what the assumed defaults are
  (pop-to-buffer simbuf)
  (goto-char (point-min))
  (insert-before-markers 
   (concat 
    "\nUsing " nrn-program-name ".\n"
    (make-string 80 ?*) "\nHOME: " (getenv "HOME")
    "\nPWD: " sim-default-directory
    "\nNEURON: " nrn-neuronhome "\nNEURONMAIN: " sim-default-directory
    "\nHOC_LIBRARY_PATH: "
    (list-to-string sim-library-path)
    "\nINITIALIZATION: " 
    nrn-init-file "\n"
    (make-string 80 ?*) "\n\n"))
  (or nrn-command-alist (make-nrn-command-alist))
  (goto-char (point-max)))

(defun nrn-index-keys ()
  "Find index files and set its key bindings, local variables, etc."
  (interactive)
  (local-set-key "\C-xl" (lambda()(interactive)(sim-load-sim t)))
  (local-set-key "\C-xf" 'sim-load-file)
  (local-set-key "\C-x?" 'sim-template-status)
  (local-set-key "\C-x " 'sim-rcs-status)
  (local-set-key "\C-x\C-c" 'sim-current-files)
  (local-set-key "\C-xc" 'hoc-neuron-command); Send command to neuron.
  (local-set-key "\C-xs" 'sim-save-sim)
  (local-set-key "\C-xz" 'sim-switch-to-buffer)
  (and window-system
       (sim-menu-bar (current-local-map))
       (nrn-menu-bar (current-local-map))))

(defun nrn-make-clearchan ()
  "Works with a list of channels."
  (interactive)
  (goto-char (point-min))
  (if (not (search-forward "Additional mechanisms from files" (point-max) t))
      (message "Unable to find additional mechanisms from files.")
    (forward-line 1)
    (let ((tmp ())
	  (files (string-to-listr 
		  " " (buffer-substring (point) (progn (end-of-line) (point))))))
      (goto-char (point-max))
      (while files
	(setq tmp (cons (concat default-directory "mod/" (car files)) tmp))
	(setq files (cdr files)))
      (setq files (list-to-string (reverse tmp)) tmp ())
      (shell-command (concat "grep " "SUFFIX " files))
      (set-buffer "*Shell Command Output*")
      (goto-char (point-min))
      (while (re-search-forward ":[ \t]*SUFFIX[ \t]+\\([a-z0-9_]+\\)" (point-max) t)
	(setq tmp (cons 
		   (buffer-substring (match-beginning 1) (match-end 1)) tmp)))
      (set-buffer (get-buffer-create "clearchans.hoc"))
      (erase-buffer)
      (insert "proc clearchans () {\nforall {\n\n")
      (while tmp
	(if (not (string= (car tmp) "nothing"))
	    (insert "  uninsert " (car tmp) "\n") )
	(setq tmp (cdr tmp)))
      (insert "\n}}\n"))
    (pop-to-buffer "clearchans.hoc")
    (hoc-mode)
    (hoc-send-proc)))

(defun nrn-update ()
  "Reset datestr and runnum known to NEURON according to current date
and emacs variable sim-runnum (which see), use prefix argument (^U) to
give specific runnum."
  (interactive)
  (let ((cbuf (current-buffer)))
    (set-buffer sim-index-buffer)
    (if current-prefix-arg
	(progn
	  (setq sim-runnum (string-to-int (read-string "sim-runnum: ")))
	  (if (or (= 0 sim-runnum) (not (integer-or-marker-p sim-runnum)))
	      (setq sim-runnum 1)))
      (setq sim-runnum 1))
    (setq sim-output-file (format "data/d%s.%02d" (date-id-string) sim-runnum))
    (comint-send-string (sim-proc) 
	(format "runnum = %d\ndatestr = \"%s\"\noutput_file = \"%s\"\nprint datestr\n"
		sim-runnum (date-id-string) sim-output-file))
    (set-buffer cbuf)))

(defun nrn-get-runnum ()   ;; set sim-runnum from the simulation
  "No longer needed since now can send a list command directly from simulator."
  (interactive)
  (setq sim-redirect-output 5
	nrn-set-var 'nrn-set-runnum) ;; the function that will do the set
  (comint-send-string (sim-proc) "\nprint runnum,\"END*&*\"\n")) ;; key symbol

(defun nrn-set-runnum (str) ;; business end to take care of string
  (set-buffer sim-index-buffer) ;; temporary switch
  (let ((rnum0 sim-runnum))
    (setq sim-runnum (string-to-int str) ;; can convert even though junk on end
	  sim-output-file (format "%s%02d" (substring sim-output-file 0 -2) sim-runnum)
	  sim-redirect-output nil)
    (if (= rnum0 (1- sim-runnum)) () ;; just update without changing index file
      (goto-char (point-max)) ;; else put in the range for autorun
      (insert (date-id-string) " autorun from run " (int-to-string rnum0) " to "
	      (int-to-string (1- sim-runnum)) "\n")
      (save-buffer))
    (set-buffer sim-buffer)))

(defun make-nrn-command-alist ()
  "Create the nrn-command-alist (which see) from procs in files from
  sim-files-alist as well as elsewhere."
  (interactive)
  (save-excursion
    (let ((comms ()) type paren hocfiles (newbuf nil)
	  (cbuf (current-buffer))
	  (sdir sim-default-directory)) ;; protect from loss by set-file
      (if (not (file-exists-p (concat nrn-src "nrnoc/hocusr.h")))
	  (message "NEURON source not found")
	(setq newbuf (set-file (concat nrn-src "nrnoc/hocusr.h")))
	(search-forward "function[] = {")
	(forward-line 1)
	(while (looking-at "\"\\([-a-z0-9_]+\\)\",")
	  (setq comms 
		(cons (cons (concat 
			     (buffer-substring (match-beginning 1) (match-end 1))
			     "()")
			    "") comms))
	  (forward-line 1))
	(if newbuf (kill-buffer nil))
	(setq newbuf (set-file (concat nrn-src "oc/hoc_init.c")))
	(while (re-search-forward "^static struct.*/\\* *\\([a-z]+\\)" (point-max) t)
	  (setq type (buffer-substring (match-beginning 1) (match-end 1)))
	  (setq paren (if (or (string= type "Keywords") (string= type "Constants"))
			  "" "()"))
	  (forward-line 4)
	  (while (not (looking-at "^}"))
	    (if (looking-at "\t\"\\([-a-z0-9_]+\\)\",")
		(setq comms 
		      (cons (cons (concat 
				   (buffer-substring (match-beginning 1) (match-end 1))
				   paren)
				  "") comms)))
	    (forward-line 1)))
	(if newbuf (kill-buffer nil)))
      (set-buffer cbuf)
      (setq hocfiles 
	    (append
	     (directory-files sim-default-directory t ".+\\.hoc$" t)
	     (directory-files (concat nrn-simctrl "../local/hoc") t ".+\\.hoc$" t)
	     (directory-files (concat nrn-simctrl "hoc") t ".+\\.hoc$" t)
	     (if (file-exists-p (concat nrn-neuronhome "share/nrn/lib/hoc"))
		 (directory-files (concat nrn-neuronhome "share/nrn/lib/hoc") t ".+\\.hoc$" t))))
      (while hocfiles
	(cond 
	 ((file-exists-p (car hocfiles))
	  (setq newbuf (set-file (car hocfiles)))
	  (while (re-search-forward "^[fp][ru][on]c *\\([-a-z0-9_]+\\) *(" (point-max) t)
	    (setq comms (cons (cons 
			       (concat
				(buffer-substring 
				 (match-beginning 1) (match-end 1)) "()")
			       (car hocfiles)) comms)))
	  (if newbuf (kill-buffer nil))))
	(setq hocfiles (cdr hocfiles)))
      (setq nrn-command-alist comms)
      )))

(defun nrn-update-command-alist (comm)
"If COMMAND is not already on nrn-command-alit, add it on."
  (or (assoc (concat comm "()") nrn-command-alist)
      (setq nrn-command-alist
	    (cons (cons (concat comm "()") "")
		  nrn-command-alist))))

(defun nrn-file2command-alist ()
  (interactive)
  (save-excursion
    (let (comm (file buffer-file-name))
      (goto-char (point-min))
      (while (re-search-forward "^[fp][ru][on]c *\\([-a-z0-9_]+\\) *(" (point-max) t)
	(setq comm (buffer-substring (match-beginning 1) (match-end 1)))
	(if (assoc (concat comm "()") nrn-command-alist) ()
	  (setq nrn-command-alist 
		(cons (cons (concat comm "()") file) nrn-command-alist)))))))

(defun nrn-append-alist ()
  "add a word to  nrn-command-alist"
  (interactive)
  (setq nrn-command-alist
	(cons (cons 
	       (buffer-substring (point) (progn (comint-bol nil)(point)))
	       "")
	      nrn-command-alist)))

(defun nrn-complete-symbol ()
  "Perform completion on partial NEURON command preceding point.  That symbol is
compared against known NEURON command that exist and any additional characters
determined by what is there are inserted."
  (interactive)
  (let* ((end (point))
	 (beg (progn (re-search-backward 
		      (concat "\\([ \t]+\\)\\|\\(" comint-prompt-regexp "\\)"))
		     (match-end 0)))
	 (pattern (buffer-substring beg end))
	 (completion (try-completion pattern nrn-command-alist nil)))
    (goto-char end)
    (cond ((eq completion t))
	  ((null completion)
	   (message "Can't find completion for \"%s\"" pattern)
	   (ding))
	  ((not (string= pattern completion))
	   (delete-region beg end)
	   (insert completion)
	   (if (string-match "()" completion) (forward-char -1)))
	  (t
	   (message "Making completion list...")
	   (let ((list (all-completions pattern nrn-command-alist nil)))
	     (with-output-to-temp-buffer " *Completions*"
	       (display-completion-list list)))
	   (message "Making completion list...%s" "done")))))

(defun nrn-load-scene ()
"Load in scene.hoc file in a crash free manner.
If load things with xopen and have an error (such as trying to graph
something that doesn't exist) will crash."
  (interactive)
  (let* ((sbuf sim-buffer)
	 (newbuf 
	  (set-file 
	   (cond ((file-exists-p (concat sim-default-directory "scene.ses"))
				 (concat sim-default-directory "scene.ses"))
		 ((file-exists-p (concat sim-default-directory "scene.hoc"))
				 (concat sim-default-directory "scene.hoc"))
		 ((expand-file-name
		   (read-file-name "Scene file: "
				   sim-default-directory "scene.ses" t ".ses")))))))
    (sim-copy-local-variables sim-localvar-list sbuf (current-buffer))
    (hoc-send-region (point-min)(point-max))
    ;; give time for neuron to start sending stuff for redirect
    (sleep-for 1.5)))

(provide 'nrn)
