;; $Id: nrnutil.el,v 1.5 2000/04/06 22:33:30 billl Exp $

;; This file is part of GNU Emacs.

;; GNU Emacs is distributed in the hope that it will be useful,
;; but WITHOUT ANY WARRANTY.  No author or distributor
;; accepts responsibility to anyone for the consequences of using it
;; or for whether it serves any particular purpose or works at all,
;; unless he says so in writing.  Refer to the GNU Emacs General Public
;; License for full details.

;; Everyone is granted permission to copy, modify and redistribute
;; GNU Emacs, but only under the conditions described in the
;; GNU Emacs General Public License.   A copy of this license is
;; supposed to have been given to you along with GNU Emacs so you
;; can know your rights and responsibilities.  It should be in a
;; file named COPYING.  Among other things, the copyright notice
;; and this notice must be preserved on all copies.
;; 
;; simctrl package by Bill Lytton - 11/1/93
;;

(defsubst reduce-file-name (filename)
  (file-relative-name filename (file-name-directory filename)))

(defun date-id-string ()
  "Produce an identifying string of form '90mar02' for file names"
  (let ((cts (current-time-string)))
    (format "%s%s%c%c" (substring cts -2) (downcase (substring cts 4 7))
	    ;; if there's a space replace with a 0
	    (if (char-equal ?  (aref cts 8)) ?0 (aref cts 8)) (aref cts 9))))

(defun time-id-string ()
  "Produce an identifying string of form 'Mar 2, 1990 1725' for file names"
  (let ((cts (current-time-string)))
    (format "%s %s, %s %s%s" (substring cts 4 7)
	    (substring cts 8 10)
	    (substring cts -4)
	    (substring cts 11 13)(substring cts 14 16))))

(defsubst list-to-string (ls)
  "Removes the parens, otherwise just use (format \"%s\" str.)"
  (if ls (substring (format "%s" ls) 1 -1) ""))

(defsubst list-to-sepstr (sep ls)
  "Place SEPARATORS between items of LIST forming a string."
  (mapconcat (lambda (x) x) ls sep))

(defun string-to-listr (sep str)
  "Use SEPARATOR (single char string) to convert STRING to list of strings"
  (let ((loc 0) (ls ()) (save (match-data)))
    ;; look for seps followed by not seps followed by optional repeated (1 sep,not seps)
    ;; followed by seps: strip off leading and trailing separators
    (if (not (string-match (concat 
     "\\" sep "*\\([^\\" sep "]+\\(\\" sep "+[^\\" sep "]+\\)*\\)" "\\" sep "*") 
			   str 0))
       (progn (store-match-data save) (list str))
      (setq str (substring str (match-beginning 1) (match-end 1)))
      ;;(and (string-match (concat "\\" sep "\\" sep) str 0)
      ;;	(error "Consecutive separators: %s" str))
      ;; search for the separator, growing the list with substrings
      (while (string-match (concat "\\" sep) str loc)
	(setq ls (cons (substring str loc (match-beginning 0)) ls)
	      loc (match-end 0)))
      ;; add in the final item since not terminated by separator
      (setq ls (cons (substring str loc) ls))
      (store-match-data save)
      (nreverse ls))))
    
(defun alist-cars (als)
  "Get all the keys for an alist (the cars)."
  (let ((out ()))
    (while als
      (setq out (append out (list (car (car als))))
	    als (cdr als)
	    ))
    out))

(defun alist-cdrs (als)
  "Get all the contents of an alist (the cdrs)."
  (let ((out ()))
    (while als
      (setq out (append out (list (cdr (car als))))
	    als (cdr als)
	    ))
    out))
	    
(defsubst set-alist-item (als item val)
 "Set ALIST ITEM to VALUE."
 (setcdr (assoc item als) val))

(defsubst get-alist-item (als item)
  "Get ALIST ITEM value"
  (cdr (assoc item als)))

(defsubst frelname (f1 f2)
  "file-relative-name with truename and expand done for you."
  (file-relative-name (file-truename (expand-file-name f1))
		      (file-truename (expand-file-name f2))))

(defun put-string-on-kill-ring (string)
  "Make STRING be the first element of the kill ring."
  (setq kill-ring (cons string kill-ring))
  (if (> (length kill-ring) kill-ring-max)
      (setcdr (nthcdr (1- kill-ring-max) kill-ring) nil))
  (setq kill-ring-yank-pointer kill-ring))

(defun set-file (file)
  "Allows you to look at a file temporarily without any of the fuss
and muss.  Returns t if this is a new buffer that should be destroyed."
  (let ((newbuf nil))
    (set-buffer (or (find-buffer-visiting file)
		    (progn (setq newbuf t) (generate-new-buffer "**temp"))))
    (if newbuf (insert-file-contents file))
    (goto-char (point-min))
    newbuf
    ))

(defun send-eof ()
  "\
   Send a literal ^D to suspend a process or subprocess.\n\
   instead of sending the interrupt to terminate the subprocess \n\
   running directly under emacs."
  (interactive)
  (message "User kill job.")
  (process-send-string (get-buffer-process (current-buffer)) "\004"))

(defun mv-output (&optional arg)
  "Move the output of a process into another window"
  (interactive "P")
  (let* ((fin (progn (point-max)(beginning-of-line)(point)))
	 (aa (buffer-substring comint-last-input-end fin))
	 (ddir default-directory))
    (delete-region comint-last-input-end fin)
    (goto-char (point-max))
    (pop-to-buffer "*output*")
    (cd ddir)
    (if (not arg) (erase-buffer) 
      (goto-char (point-max)) (insert "\n"))
    (insert aa)
    (goto-char (point-min))))

(provide 'nrnutil)

