;; $Id: nrnvec.el,v 1.14 2000/04/06 22:33:30 billl Exp $

(defvar nrnvec-attrnum 0
  "distinguish the different files")
(defvar nrnvec-files ()
  "a list of files being worked on")
(setq nrnvec-grep-regexp "'^//[:pbCM ]'")
;; rebinding
(define-key hoc-mode-map [C-return] 'nrnvec-send-line-and-save)

(defun nrnvec-toggle-key ()
  "Change C-return btwn hoc-send and nrnvec-send."
  (interactive)
  (cond
   ((eq (key-binding [C-return]) 'nrnvec-send-line-and-save)
    (define-key hoc-mode-map [C-return] 'hoc-send-line-and-save))
   ((eq (key-binding [C-return]) 'hoc-send-line-and-save)
    (define-key hoc-mode-map [C-return] 'nrnvec-send-line-and-save))))

(defun nrnvec-grep-files (dir rxp)
  "Go through a set of files and run grep on each one."
  (interactive "DDir: \nsFile Regexp: ")
  (setq nrnvec-files (directory-files dir t rxp t))
  (let ((fls nrnvec-files))
    (with-output-to-temp-buffer "Files" 
      (princ (list-to-sepstr "\n" nrnvec-files)))
    (if (y-or-n-p "OK ")
	(while fls
	  (nrnvec-grep (car fls))
	  (setq fls (cdr fls)))
      (message "Aborted."))))

(defun nrnvec-load-files (dir rxp)
  "Go through a set of files and run grep on each one."
  (interactive "DDir: \nsFile Regexp: ")
  (or (string= rxp "")
      (setq nrnvec-files (directory-files dir t rxp t)))
  (let ((fls nrnvec-files) fdot fxtp (n 1)
	(dir (file-name-directory (car nrnvec-files))))
    (with-output-to-temp-buffer "Files" 
      (princ (list-to-sepstr "\n" nrnvec-files)))
    (if (y-or-n-p "OK ")
	(while fls
	  (setq fdot (concat "." (file-name-nondirectory (car fls)))
		fxtp (concat dir fdot ".xtp")
		fdot (concat dir fdot))
	  (cond
	   ((file-exists-p fdot)(nrnvec-read fdot n)(message "Reading %s" fdot))
	   ((file-exists-p fxtp)(nrnvec-read fxtp n)(message "Reading %s" fxtp))
	   (t (message "Unable to find %s or %s. Creating %s..." fdot fxtp fxtp)
 	      (nrnvec-grep (car fls))))
	  (setq fls (cdr fls) n (1+ n)))
      (message "Aborted."))))

(defun nrnvec-grep (fl)
  "Run grep on a single file to look for the line numbers."
  (interactive "fFile: ")
  (let ((fdot (concat (file-name-directory fl) "." (file-name-nondirectory fl) ".xtp")))
    (and (file-exists-p fdot) (delete-file fdot))
    (start-process-shell-command 
     "vgrp" "*Messages*" "grep" nrnvec-grep-regexp fl ">" fdot)))

(defun nrnvec-read (fdot attrnum)
  "Read vector file into hoc mode so that the individual lines can be interpreted for display."
  (interactive "fVFile: \nnAttrnum: ")
  (string-match "\\(v.+[0-9]+\\)" (file-name-nondirectory fdot))
  (let ((f (match-string 1 (file-name-nondirectory fdot)))
	cm pstep segs entries) 
    (find-file fdot)
    (goto-char (point-min))
    (re-search-forward "//: \\(.+\\)" nil t)
    (setq cm (match-string 1))
    (goto-char (point-min))
    (re-search-forward "//printStep \\(.+\\)" nil t)
    (setq pstep (match-string 1))
    (hoc-mode)
    (setq buffer-read-only t)
    (make-local-variable 'nrnvec-attrnum)
    (setq nrnvec-attrnum attrnum)
    (hoc-sender 
     (sim-proc) 
     (concat "{attrlist(5,"
	     nrnvec-attrnum ","
	     "\"" (expand-file-name f) "\","
	     "\"" cm "\","
	     pstep
             ")}\n")
     )))
	     
(defun nrnvec-sort ()
  "Change a .xtp file (raw grep output from nrnvec-grep) with multiple segs to a .v file"
  (interactive)
  (setq buffer-read-only nil)
  (goto-char (point-min))
  (re-search-forward "^//b")
  (beginning-of-line)
  (let ((top (point)) beg end str len)
    (goto-char (point-max))
    (re-search-backward "^//b")
    (forward-line 1)
    (setq end (point))
    (narrow-to-region top end)
    (sort-lines nil top end)
    (goto-char (point-min))
    (while (re-search-forward "^//b[0-9] \\([^ ]+\\) [0-9]+ \\([0-9]+\\).*" nil t)
      (setq beg (match-beginning 0)
	    str (match-string 1)
	    len (+ (string-to-int (match-string 2))
		   (1+ (- (match-end 0)(match-beginning 0))))) 
      (replace-match "\\1")
      (insert (format " %d" len))
      (while (search-forward str nil t)
	(re-search-forward "[0-9]+ \\([0-9]+\\).*")
	(setq len (+ (string-to-int (match-string 1))
		     (1+ (- (prog2 (end-of-line)(point)(beginning-of-line))(point)))))
	(beginning-of-line)
	(delete-region (point) (match-end 0))
	(insert (format "%d" len)))
      (setq end (point))
      (sort-numeric-fields 1 beg end)
      (goto-char beg)
      (while (search-forward "\n" end t)(replace-match " ")))
    (sort-numeric-fields 2 (point-min) (point-max))
    (widen)
    (goto-char (point-min))
    (setq buffer-read-only t)
    (write-file (substring buffer-file-name 0 -4) t)))

;; overwrite the binding in hoc.el
(defun nrnvec-send-line-and-save ()
  "Screens for //b line; otherwise like hoc-send-line-and-save except
only does one buffer"
  (interactive)
  (setq sim-redirect-output nil)
  (save-excursion
    (let ((cbuf (current-buffer))
	  comm 
	  (eolptr (progn (end-of-line) (point)))) ; Mark end of line.
      (beginning-of-line)		; Move to first character.
      (setq comm (concat (buffer-substring (point) eolptr) "\n"))
      (cond
       ((string-match "\\`\\([^ ]+\\) \\([0-9]+ [0-9]+ \\([0-9]+ \\)+\\)" comm)
	(setq 
	 comm 
	 (concat "{rvl(" nrnvec-attrnum ",\""
		 (match-string 1 comm) "\","
		 (list-to-sepstr "," (string-to-listr " "(substring (match-string 2 comm) 0 -1)))
		 ")}\n")))
       ((string-match "\\`//b[0-9] \\([^ ]+\\) [0-9]+ \\([0-9]+\\)" comm)
	(setq comm (concat "{rvl(" nrnvec-attrnum ",\"" 
		 (match-string 1 comm) "\","
		 (match-string 2 comm) ")}\n"))))
      (save-buffer)
      (if hoc-send-all-sim-buffers
	  (mapcar (lambda (x) (set-buffer(process-buffer x))(insert-before-markers comm))
		  (sim-all-processes))
	(set-buffer sim-buffer)
	(insert-before-markers comm))
      (ring-insert comint-input-ring (substring comm 0 (1- (length comm))))
      (set-buffer cbuf)
      (hoc-sender (sim-proc) comm)
      ))
  (forward-line 1))

(defvar nrnvec-comm "vlk" "command to run with printlist vecs; form
should be eg either vlk or .printf")
(defun nrnvec-setcomm ()
  "command to run with printlist vecs; form
should be eg either vlk or .printf"
  (interactive)
  (setq nrnvec-comm 
	(completing-read "Comm: " 
			 '(("vlk")(".printf")(".min")(".max")(".size")
			   ("vpr")
			  ))))

(defun nrnvec-printlist ()
  "takes an entry of form 'num varname' and does something with vec;
use: for ltr(XO,printlist) { printf(\"%d %s\n\",i1,XO.var) }"
  (interactive)
  (setq sim-redirect-output nil)
  (save-excursion
    (let ((cbuf (current-buffer))
	  comm num
	  (eolptr (progn (end-of-line) (point)))) ; Mark end of line.
      (beginning-of-line)		; Move to first character.
      (re-search-forward "\\([0-9]+\\) +\\([^ ]+\\)")
      (setq 
       num (string-to-int (match-string 1))
       comm 
       (if (string-match "^\\." nrnvec-comm)
	   (format "printlist.object(%d).vec%s\n" num nrnvec-comm)
	 (format "%s(printlist.object(%d).vec)\n" 
		 nrnvec-comm num)))
      (set-buffer sim-buffer)
      (insert-before-markers comm)
      (ring-insert comint-input-ring (substring comm 0 (1- (length comm))))
      (set-buffer cbuf)
      (hoc-sender (sim-proc) comm)
      ))
  (forward-line 1))

